""" Plugin for importing data from spansh """
from __future__ import annotations

from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime, timedelta
from email.utils import parsedate_to_datetime
from pathlib import Path
from rich.progress import Progress

from .. import plugins, cache, transfers, csvexport, corrections

import os
import requests
import sqlite3
import sys
import time
import typing
import ijson

if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from dataclasses import dataclass
else:
    dataclass = False  # pylint: disable=invalid-name

if typing.TYPE_CHECKING:
    from typing import Any, Optional
    from collections.abc import Iterable
    from .. tradeenv import TradeEnv

SOURCE_URL = 'https://downloads.spansh.co.uk/galaxy_stations.json'

STATION_TYPE_MAP = {
    'None': [0, False],
    'Outpost': [1, False],
    'Coriolis Starport': [2, False],
    'Ocellus Starport': [3, False],
    'Orbis Starport': [4, False],
    'Planetary Outpost': [11, True],
    'Planetary Port': [12, True],
    'Mega ship': [13, False],
    'Asteroid base': [14, False],
    'Drake-Class Carrier': [24, False],  # fleet carriers
    'Settlement': [25, True],            # odyssey settlements
}

if dataclass:
    # Dataclass with slots is considerably cheaper and faster than namedtuple
    # but is only reliably introduced in 3.10+
    @dataclass(slots=True)
    class System:
        id:             int
        name:           str
        pos_x:          float
        pos_y:          float
        pos_z:          float
        modified:       float | None
    
    @dataclass(slots=True)
    class Station:      # pylint: disable=too-many-instance-attributes
        id:             int
        system_id:      int
        name:           str
        distance:       float
        max_pad_size:   str
        market:         str     # should be Optional[bool]
        black_market:   str     # should be Optional[bool]
        shipyard:       str     # should be Optional[bool]
        outfitting:     str     # should be Optional[bool]
        rearm:          str     # should be Optional[bool]
        refuel:         str     # should be Optional[bool]
        repair:         str     # should be Optional[bool]
        planetary:      str     # should be Optional[bool]
        type:           int     # station type
        modified:       float
    
    @dataclass(slots=True)
    class Ship:
        id:             int
        name:           str
        modified:       float
    
    @dataclass(slots=True)
    class Module:
        id:             int
        name:           str
        cls:            int
        rating:         str
        ship:           str
        modified:       float
    
    @dataclass(slots=True)
    class Commodity:
        id:             int
        name:           str
        category:       str
        demand:         int
        supply:         int
        sell:           int
        buy:            int
        modified:       float

else:
    System = namedtuple('System', 'id,name,pos_x,pos_y,pos_z,modified')
    Station = namedtuple('Station',
                         'id,system_id,name,distance,max_pad_size,'
                         'market,black_market,shipyard,outfitting,rearm,refuel,repair,planetary,type,modified')
    Ship = namedtuple('Ship', 'id,name,modified')
    Module = namedtuple('Module', 'id,name,cls,rating,ship,modified')
    Commodity = namedtuple('Commodity', 'id,name,category,demand,supply,sell,buy,modified')


class Timing:
    """ Helper that provides a context manager for timing code execution. """
    
    def __init__(self):
        self.start_ts = None
        self.end_ts = None
    
    def __enter__(self):
        self.start_ts = time.perf_counter()
        self.end_ts = None
        return self
    
    def __exit__(self, *args):
        self.end_ts = time.perf_counter()
    
    @property
    def elapsed(self) -> Optional[float]:
        """ If the timing has finished, calculates the elapsed time. """
        if self.start_ts is None:
            return None
        return (self.end_ts or time.perf_counter()) - self.start_ts
    
    @property
    def is_finished(self) -> bool:
        """ True if the timing has finished. """
        return self.end_ts is not None


class Progresser:
    """ Encapsulates a potentially transient progress view for a given TradeEnv. """
    def __init__(self, tdenv: 'TradeEnv', title: str, fancy: bool = True, total: Optional[int] = None):
        self.started = time.time()
        self.tdenv = tdenv
        self.progress, self.main_task = None, None
        self.title = title
        self.fancy = fancy
        self.total = total
        self.main_task = None
        if fancy:
            self.progress = Progress(console=self.tdenv.console, transient=True, auto_refresh=True, refresh_per_second=2)
        else:
            self.progress = None
    
    def __enter__(self):
        if not self.fancy:
            self.tdenv.uprint(self.title)
        else:
            self.progress.start()
            self.main_task = self.progress.add_task(self.title, start=True, total=self.total)
        return self
    
    def __exit__(self, *args):
        self.progress.stop()
    
    def update(self, title: str) -> None:
        if self.fancy:
            self.progress.update(self.main_task, description=title)
        else:
            self.tdenv.DEBUG1(title)
    
    @contextmanager
    def task(self, title: str, total: Optional[int] = None, parent: Optional[str] = None):
        parent = parent or self.main_task
        if self.fancy:
            task = self.progress.add_task(title, start=True, total=total, parent=parent)
        else:
            self.tdenv.DEBUG0(title)
            task = None
        try:
            yield task
        finally:
            if self.fancy:
                self.progress.remove_task(task)
        if task is not None and parent is not None:
            self.progress.update(parent, advance=1)
    
    def bump(self, task, advance: int = 1, description: Optional[str] = None):
        """ Advances the progress of a task by one mark. """
        if self.fancy and task is not None:
            self.progress.update(task, advance=advance, description=description)


def get_timings(started: float, system_count: int, total_station_count: int, *, min_count: int = 100) -> tuple[float, str]:
    """ describes how long it is taking to process each system and station """
    elapsed = time.time() - started
    timings = "sys="
    if system_count >= min_count:
        avg = elapsed / float(system_count) * 1000.0
        timings += f"{avg:5.2f}ms"
    else:
        timings += "..."
    timings += ", stn="
    if total_station_count >= min_count:
        avg = elapsed / float(total_station_count) * 1000.0
        timings += f"{avg:5.2f}ms"
    else:
        timings += "..."
    return elapsed, timings


class ImportPlugin(plugins.ImportPluginBase):
    """Plugin that downloads data from https://spansh.co.uk/dumps.
    """
    
    pluginOptions = {
        'url': f'URL to download galaxy data from (defaults to {SOURCE_URL})',
        'file': 'Local filename to import galaxy data from; use "-" to load from stdin',
        'maxage': 'Skip all entries older than specified age in days, ex.: maxage=1.5',
    }
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = self.getOption('url')
        self.file = self.getOption('file')
        self.maxage = float(self.getOption('maxage')) if self.getOption('maxage') else None
        assert not (self.url and self.file), 'Provide either url or file, not both'
        if self.file and (self.file != '-'):
            self.file = (Path(self.tdenv.cwDir, self.file)).resolve()
        if not Path(self.tdb.dataPath, "TradeDangerous.db").exists():
            ri_path = Path(self.tdb.dataPath, "RareItem.csv")
            rib_path = ri_path.with_suffix(".tmp")
            if ri_path.exists():
                if rib_path.exists():
                    rib_path.unlink()
                ri_path.rename(rib_path)
            cache.buildCache(self.tdb, self.tdenv)
            if ri_path.exists():
                ri_path.unlink()
            if rib_path.exists():
                rib_path.rename(ri_path)
        
        self.need_commit = False
        self.cursor = self.tdb.getDB().cursor()
        self.commit_rate = 200
        self.commit_limit = self.commit_rate
        
        self.known_systems = self.load_known_systems()
        self.known_stations = self.load_known_stations()
        self.known_ships = self.load_known_ships()
        self.known_modules = self.load_known_modules()
        self.known_commodities = self.load_known_commodities()
    
    def print(self, *args, **kwargs) -> None:
        """ Shortcut to the TradeEnv uprint method. """
        self.tdenv.uprint(*args, **kwargs)
    
    def commit(self, *, force: bool = False) -> None:
        """ Perform a commit if required, but try not to do a crazy amount of committing. """
        if not force and not self.need_commit:
            return
        
        if not force and self.commit_limit > 0:
            self.commit_limit -= 1
            return
        
        db = self.tdb.getDB()
        db.commit()
        self.cursor = db.cursor()
        
        self.commit_limit = self.commit_rate
        self.need_commit = False
    
    def run(self):
        if not self.tdenv.detail:
            self.print('This will take at least several minutes...')
            self.print('You can increase verbosity (-v) to get a sense of progress')
        
        theme = self.tdenv.theme
        BOLD, CLOSE, DIM, ITALIC = theme.bold, theme.CLOSE, theme.dim, theme.italic  # pylint: disable=invalid-name
        if not self.file:
            url = self.url or SOURCE_URL
            local_mod_time = 0
            self.file = Path(self.tdenv.tmpDir, "galaxy_stations.json")
            if self.file.exists():
                local_mod_time = self.file.stat().st_mtime
            
            headers = {"User-Agent": "Trade-Dangerous", "Accept-Encoding": "identity"}
            try:
                response = requests.head(url, headers=headers, timeout=70)
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.tdenv.WARN("Problem with download:\n    URL: {}\n    Error: {}", url, str(e))
                return False
            last_modified = response.headers.get("last-modified")
            dump_mod_time = parsedate_to_datetime(last_modified).timestamp()
            if local_mod_time < dump_mod_time:
                if self.file.exists():
                    self.file.unlink()
                self.print(f'Downloading prices from remote URL: {url}')
                try:
                    transfers.download(self.tdenv, url, self.file)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    self.tdenv.WARN("Problem with download:\n    URL: {}\n    Error: {}", url, str(e))
                    return False
                self.print(f'Download complete, saved to local file: "{self.file}"')
                os.utime(self.file, (dump_mod_time, dump_mod_time))
        
        sys_desc = f"Importing {ITALIC}spansh{CLOSE} data"
        
        # TODO: find a better way to get the total number of systems
        # A bad way to do it:
        total_systems = 0
        if self.tdenv.detail:
            print('Counting total number of systems...')
        with open(self.file, 'r', encoding='utf8') as stream:
            for system_data in ijson.items(stream, 'item', use_float=True):
                total_systems += 1
                if (not total_systems % 250) and self.tdenv.detail:
                    print(f'Total systems: {total_systems}', end='\r')
        
        if self.tdenv.detail:
            print(f'Total systems: {total_systems}')
        
        with Timing() as timing, Progresser(self.tdenv, sys_desc, total=total_systems) as progress:
        # with Timing() as timing, Progresser(self.tdenv, sys_desc, total=len(self.known_stations)) as progress:
            system_count = 0
            total_station_count = 0
            total_ship_count = 0
            total_module_count = 0
            total_commodity_count = 0
            
            age_cutoff = timedelta(days=self.maxage) if self.maxage else None
            now = datetime.now()
            started = time.time()
            
            for system, station_iter in self.data_stream():
                upper_sys = system.name.upper()
                
                elapsed, averages = get_timings(started, system_count, total_station_count)
                label = f"{ITALIC}#{system_count:<5d}{CLOSE} {BOLD}{upper_sys:30s}{CLOSE} {DIM}({elapsed:.2f}s, avgs: {averages}){CLOSE}"
                stations = list(station_iter)
                with progress.task(label, total=len(stations)) as sta_task:
                    if system.id not in self.known_systems:
                        self.ensure_system(system, upper_sys)
                    
                    station_count = 0
                    ship_count = 0
                    module_count = 0
                    commodity_count = 0
                    
                    for station, ships, modules, commodities in stations:
                        fq_station_name = f'@{upper_sys}/{station.name}'
                        
                        station_info = self.known_stations.get(station.id)
                        if not station_info or station.modified > station_info[2]:
                            self.ensure_station(station)
                        elif station_info[1] != station.system_id:
                            self.print(f'        |  {station.name:50s}  |  Megaship station moved, updating system')
                            self.execute("UPDATE Station SET system_id = ? WHERE station_id = ?", station.system_id, station.id, commitable=True)
                            self.known_stations[station.id] = (station.name, station.system_id, station.modified)
                        
                        # Ships
                        ship_entries = []
                        db_ship_times = dict(self.execute("SELECT ship_id, modified FROM ShipVendor WHERE station_id = ?", station.id))
                        
                        for ship in ships:
                            if ship.id not in self.known_ships:
                                ship = self.ensure_ship(ship)
                            
                            # We're concerned with the ship age, not the station age,
                            # as they each have their own 'modified' times.
                            if age_cutoff and (now - ship.modified) > age_cutoff:
                                if self.tdenv.detail:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping shipyard due to age: {now - ship.modified}, ts: {ship.modified}')
                                break
                            db_modified = db_ship_times.get(ship.id)
                            modified = parse_ts(db_modified) if db_modified else None
                            if modified and ship.modified <= modified:
                                # All ships in a station will have the same modified time,
                                # so no need to check the rest if the first is older.
                                if self.tdenv.detail > 2:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping older shipyard data')
                                break
                            
                            ship_entries.append((ship.id, station.id, ship.modified))
                        if ship_entries:
                            self.executemany("""INSERT OR REPLACE INTO ShipVendor (
                                ship_id, station_id, modified
                            ) VALUES (
                                ?, ?, IFNULL(?, CURRENT_TIMESTAMP)
                            )""", ship_entries, commitable=True)
                            ship_count += len(ship_entries)
                        
                        # Upgrades
                        module_entries = []
                        db_module_times = dict(self.execute("SELECT upgrade_id, modified FROM UpgradeVendor WHERE station_id = ?", station.id))
                        
                        for module in modules:
                            if module.id not in self.known_modules:
                                module = self.ensure_module(module)
                            
                            # We're concerned with the outfitting age, not the station age,
                            # as they each have their own 'modified' times.
                            if age_cutoff and (now - module.modified) > age_cutoff:
                                if self.tdenv.detail:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping outfitting due to age: {now - station.modified}, ts: {station.modified}')
                                break
                            db_modified = db_module_times.get(module.id)
                            modified = parse_ts(db_modified) if db_modified else None
                            if modified and module.modified <= modified:
                                # All modules in a station will have the same modified time,
                                # so no need to check the rest if the fist is older.
                                if self.tdenv.detail > 2:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping older outfitting data')
                                break
                            
                            module_entries.append((module.id, station.id, module.modified))
                        if module_entries:
                            self.executemany("""INSERT OR REPLACE INTO UpgradeVendor (
                                upgrade_id, station_id, modified
                            ) VALUES (
                                ?, ?, IFNULL(?, CURRENT_TIMESTAMP)
                            )""", module_entries, commitable=True)
                            module_count += len(module_entries)
                        
                        # Items
                        commodity_entries = []
                        db_commodity_times = dict(self.execute("SELECT item_id, modified FROM StationItem WHERE station_id = ?", station.id))
                        
                        for commodity in commodities:
                            if commodity.id not in self.known_commodities:
                                commodity = self.ensure_commodity(commodity)
                            
                            # We're concerned with the market age, not the station age,
                            # as they each have their own 'modified' times.
                            if age_cutoff and (now - commodity.modified) > age_cutoff:
                                if self.tdenv.detail:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping market due to age: {now - station.modified}, ts: {station.modified}')
                                break
                            
                            db_modified = db_commodity_times.get(commodity.id)
                            modified = parse_ts(db_modified) if db_modified else None
                            if modified and commodity.modified <= modified:
                                # All commodities in a station will have the same modified time,
                                # so no need to check the rest if the fist is older.
                                if self.tdenv.detail > 2:
                                    self.print(f'        |  {fq_station_name:50s}  |  Skipping older market data')
                                break
                            commodity_entries.append((station.id, commodity.id, commodity.modified,
                                                      commodity.sell, commodity.demand, -1,
                                                      commodity.buy, commodity.supply, -1, 0))
                        if commodity_entries:
                            self.executemany("""INSERT OR REPLACE INTO StationItem (
                                station_id, item_id, modified,
                                demand_price, demand_units, demand_level,
                                supply_price, supply_units, supply_level, from_live
                            ) VALUES (
                                ?, ?, IFNULL(?, CURRENT_TIMESTAMP),
                                ?, ?, ?,
                                ?, ?, ?, ?
                            )""", commodity_entries, commitable=True)
                            commodity_count += len(commodity_entries)
                        # Good time to save data and try to keep the transaction small
                        self.commit()
                        
                        if commodity_count or ship_count or module_count:
                            station_count += 1
                        progress.bump(sta_task)
                
                system_count += 1
                if station_count:
                    total_station_count += station_count
                    total_ship_count += ship_count
                    total_module_count += module_count
                    total_commodity_count += commodity_count
                    if self.tdenv.detail:
                        self.print(
                            f'{system_count:6d}  |  {upper_sys:50s}  |  '
                            f'{station_count:3d} st {commodity_count:5d} co '
                            f'{ship_count:4d} sh {module_count:4d} mo'
                        )
                self.commit()
                
                if not system_count % 25:
                    avg_stations = total_station_count / (system_count or 1)
                    progress.update(f"{sys_desc}{DIM} ({total_station_count}:station:, {system_count}:glowing_star:, {avg_stations:.1f}:station:/:glowing_star:){CLOSE}")
            
            self.commit()
            self.tdb.close()
            self.print(
                f'{timedelta(seconds=int(timing.elapsed))!s}  Done  '
                f'{total_station_count} st {total_commodity_count} co '
                f'{total_ship_count} sh {total_module_count} mo'
            )
        
        with Timing() as timing:
            # Need to make sure cached tables are updated
            self.print('Exporting to cache...')
            for table in ("Item", "Station", "System", "StationItem", "Ship", "ShipVendor", "Upgrade", "UpgradeVendor"):
                self.print(f'Exporting {table}.csv            ', end='\r')
                csvexport.exportTableToFile(self.tdb, self.tdenv, table)
            self.print('Exporting TradeDangerous.prices', end='\r')
            cache.regeneratePricesFile(self.tdb, self.tdenv)
            self.print(f'Cache export completed in {timedelta(seconds=int(timing.elapsed))!s}')
        
        return False
    
    def data_stream(self):
        stream = None
        if self.file == '-':
            self.print('Reading data from stdin')
            stream = sys.stdin
        elif self.file:
            self.print(f'Reading data from local file: "{self.file}"')
            stream = open(self.file, 'r', encoding='utf8')
        return self.ingest_stream(stream)
    
    def execute(self, query: str, *params, commitable: bool = False) -> sqlite3.Cursor:
        """ helper method that performs retriable queries and marks the transaction 
            as needing to commit if the query is commitable."""
        if commitable:
            self.need_commit = True
        attempts = 5
        while True:
            try:
                return self.cursor.execute(query, params)
            except sqlite3.OperationalError as ex:
                if "no transaction is active" in str(ex):
                    self.print(f"no transaction for {query}")
                    raise
                if not attempts:
                    raise
                attempts -= 1
                self.print(f'Retrying query \'{query}\': {ex!s}')
                time.sleep(1)
    
    def executemany(self, query: str, data: Iterable[Any], *, commitable: bool = False) -> sqlite3.Cursor:
        """ helper method that performs retriable queries and marks the transaction as needing to commit
            if the query is commitable."""
        if commitable:
            self.need_commit = True
        attempts = 5
        while True:
            try:
                return self.cursor.executemany(query, data)
            except sqlite3.OperationalError as ex:
                if "no transaction is active" in str(ex):
                    self.print(f"no transaction for {query}")
                    raise
                if not attempts:
                    raise
                attempts -= 1
                self.print(f'Retrying query \'{query}\': {ex!s}')
                time.sleep(1)
    
    def load_known_systems(self) -> dict[int, str]:
        """ Returns a dictionary of {system_id -> system_name} for all current systems in the database. """
        try:
            return dict(self.cursor.execute('SELECT system_id, name FROM System'))
        except Exception as e:  # pylint: disable=broad-except
            self.print("[purple]:thinking_face:Assuming no system data yet")
            self.tdenv.DEBUG0(f"load_known_systems query raised {e}")
            return {}
    
    def load_known_stations(self) -> dict[int, tuple[str, int, float]]:
        """ Returns a dictionary of {station_id -> (station_name, system_id, modified)} for all current stations in the database. """
        try:
            return {cols[0]: (cols[1], cols[2], parse_ts(cols[3])) for cols in self.cursor.execute('SELECT station_id, name, system_id, modified FROM Station')}
        except Exception as e:  # pylint: disable=broad-except
            self.print("[purple]:thinking_face:Assuming no station data yet")
            self.tdenv.DEBUG0(f"load_known_stations query raised {e}")
            return {}
    
    def load_known_ships(self):
        """ Returns a dictionary of {ship_id -> name} for all current ships in the database. """
        try:
            return dict(self.cursor.execute('SELECT ship_id, name FROM Ship'))
        except Exception as e:  # pylint: disable=broad-except
            self.print("[purple]:thinking_face:Assuming no ship data yet")
            self.tdenv.DEBUG0(f"load_known_ships query raised {e}")
            return {}
    
    def load_known_modules(self):
        """ Returns a dictionary of {upgrade_id -> name} for all current modules in the database. """
        try:
            return dict(self.cursor.execute('SELECT upgrade_id, name FROM Upgrade'))
        except Exception as e:  # pylint: disable=broad-except
            self.print("[purple]:thinking_face:Assuming no module data yet")
            self.tdenv.DEBUG0(f"load_known_modules query raised {e}")
            return {}
    
    def load_known_commodities(self):
        """ Returns a dictionary of {fdev_id -> name} for all current commodities in the database. """
        try:
            return dict(self.cursor.execute('SELECT fdev_id, name FROM Item'))
        except Exception as e:  # pylint: disable=broad-except
            self.print("[purple]:thinking_face:Assuming no commodity data yet")
            self.tdenv.DEBUG0(f"load_known_commodities query raised {e}")
            return {}
    
    def ensure_system(self, system: System, upper_name: str) -> None:
        """ Adds a record for a system, and registers the system in the known_systems dict. """
        self.execute(
            '''
            INSERT INTO System (system_id, name, pos_x, pos_y, pos_z, modified) VALUES (?, ?, ?, ?, ?, ?)
            ''',
            system.id, system.name, system.pos_x, system.pos_y, system.pos_z, system.modified,
            commitable=True,
        )
        if self.tdenv.detail > 1:
            self.print(f'        |  {upper_name:50s}  |  Added missing system :glowing_star:')
        self.known_systems[system.id] = system.name
    
    def ensure_station(self, station: Station) -> None:
        """ Adds a record for a station, and registers the station in the known_stations dict. """
        self.execute(
            '''
            INSERT OR REPLACE INTO Station (
                system_id, station_id, name,
                ls_from_star, max_pad_size,
                market, blackmarket, shipyard, outfitting,
                rearm, refuel, repair,
                planetary,
                modified,
                type_id
            )
            VALUES (
                ?, ?, ?,
                ?, ?,
                ?, ?, ?, ?,
                ?, ?, ?,
                ?,
                ?,
                ?
            )
            ''',
            station.system_id,
            station.id,
            station.name,
            station.distance,
            station.max_pad_size,
            self.bool_yn(station.market),
            self.bool_yn(station.black_market),
            self.bool_yn(station.shipyard),
            self.bool_yn(station.outfitting),
            self.bool_yn(station.rearm),
            self.bool_yn(station.refuel),
            self.bool_yn(station.repair),
            self.bool_yn(station.planetary),
            station.modified,
            station.type,
            commitable=True,
        )
        note = "Updated" if self.known_stations.get(station.id) else "Added"
        if self.tdenv.detail > 1:
            system_name = self.known_systems[station.system_id]
            upper_sys = system_name.upper()
            fq_station_name = f'@{upper_sys}/{station.name}'
            self.print(f'        |  {fq_station_name:50s}  |  {note} station')
        self.known_stations[station.id] = (station.name, station.system_id, station.modified)
    
    def ensure_ship(self, ship: Ship):
        """ Adds a record for a ship, and registers the ship in the known_ships dict. """
        self.execute(
            '''
            INSERT INTO Ship (ship_id, name) VALUES (?, ?)
            ''',
            ship.id, ship.name,
            commitable=True,
        )
        self.known_ships[ship.id] = ship.name
        
        return ship
    
    def ensure_module(self, module: Module):
        """ Adds a record for a module, and registers the module in the known_modules dict. """
        self.execute(
            '''
            INSERT INTO Upgrade (upgrade_id, name, class, rating, ship) VALUES (?, ?, ?, ?, ?)
            ''',
            module.id, module.name, module.cls, module.rating, module.ship,
            commitable=True,
        )
        self.known_modules[module.id] = module.name
        
        return module
    
    def ensure_commodity(self, commodity: Commodity):
        """ Adds a record for a commodity and registers the commodity in the known_commodities dict. """
        self.execute(
            '''
            INSERT INTO Item (item_id, category_id, name, fdev_id)
            VALUES (?, (SELECT category_id FROM Category WHERE upper(name) = ?), ?, ?)
            ''',
            commodity.id,
            commodity.category.upper(),
            corrections.correctItem(commodity.name),
            commodity.id,
            commitable=True,
        )
        
        # Need to update ui_order
        temp = self.execute("""SELECT name, category_id, fdev_id, ui_order
                        FROM Item
                        ORDER BY category_id, name
                        """)
        cat_id = 0
        ui_order = 1
        self.tdenv.DEBUG0("Updating ui_order data for items.")
        changes = []
        for name, db_cat, fdev_id, db_order in temp:
            if db_cat != cat_id:
                ui_order = 1
                cat_id = db_cat
            else:
                ui_order += 1
            if ui_order != db_order:
                self.tdenv.DEBUG0(f"UI order for {name} ({fdev_id}) needs correction.")
                changes += [(ui_order, fdev_id)]
        
        if changes:
            self.executemany(
                "UPDATE Item SET ui_order = ? WHERE fdev_id = ?",
                changes,
                commitable=True
            )
        
        self.known_commodities[commodity.id] = commodity.name
        
        return commodity
    
    def bool_yn(self, value: Optional[bool]) -> str:
        """ translates a ternary (none, true, false) into the ?/Y/N representation """
        return '?' if value is None else ('Y' if value else 'N')
    
    def ingest_stream(self, stream):
        """Ingest a spansh-style galaxy dump, yielding system-level data."""
        for system_data in ijson.items(stream, 'item', use_float=True):
            if "Shinrarta Dezhra" in system_data.get('name') and self.tdenv.debug:
                with open(Path(self.tdenv.tmpDir, "shin_dez.json"), 'w') as file:
                    # file.write(system_data)
                    import json
                    json.dump(system_data, file, indent=4)
            
            coords = system_data.get('coords', {})
            yield (
                System(
                    id=system_data.get('id64'),
                    name=system_data.get('name', 'Unnamed').strip(),
                    pos_x=coords.get('x', 999999),
                    pos_y=coords.get('y', 999999),
                    pos_z=coords.get('z', 999999),
                    modified=parse_ts(system_data.get('date')),
                ),
                ingest_stations(system_data),
            )


def ingest_stations(system_data):
    """Ingest system-level data, yielding station-level data."""
    sys_id = system_data.get('id64')
    targets = [system_data, *system_data.get('bodies', ())]
    for target in targets:
        for station_data in target.get('stations', ()):
            services = set(station_data.get('services', ()))
            shipyard = None
            if 'Shipyard' in services:
                shipyard = station_data.get('shipyard', {})
            outfitting = None
            if 'Outfitting' in services:
                outfitting = station_data.get('outfitting', {})
            market = None
            if 'Market' in services:
                market = station_data.get('market', {})
            if not shipyard and not outfitting and not market:
                continue
            landing_pads = station_data.get('landingPads', {})
            max_pad_size = '?'
            if landing_pads.get('large'):
                max_pad_size = 'L'
            elif landing_pads.get('medium'):
                max_pad_size = 'M'
            elif landing_pads.get('small'):
                max_pad_size = 'S'
            station_type = STATION_TYPE_MAP.get(station_data.get('type'))
            yield (
                Station(
                    id=station_data.get('id'),
                    system_id=sys_id,
                    name=station_data.get('name', 'Unnamed').strip(),
                    distance=station_data.get('distanceToArrival', 999999),
                    max_pad_size=max_pad_size,
                    market='Market' in services,
                    black_market='Black Market' in services,
                    shipyard='Shipyard' in services,
                    outfitting='Outfitting' in services,
                    rearm='Restock' in services,
                    refuel='Refuel' in services,
                    repair='Repair' in services,
                    planetary=station_type[1] if station_type else False,
                    type=station_type[0] if station_type else 0,
                    modified=parse_ts(station_data.get('updateTime')),
                ),
                ingest_shipyard(shipyard),
                ingest_outfitting(outfitting),
                ingest_market(market),
            )

def ingest_shipyard(shipyard):
    """Ingest station-level market data, yielding commodities."""
    if not shipyard or not shipyard.get('ships'):
        return None
    for ship in shipyard['ships']:
        yield Ship(
            id=ship.get('shipId'),
            name=ship.get('name'),
            modified=parse_ts(shipyard.get('updateTime'))
        )

def ingest_outfitting(outfitting):
    """Ingest station-level market data, yielding commodities."""
    if not outfitting or not outfitting.get('modules'):
        return None
    for module in outfitting['modules']:
        yield Module(
            id=module.get('moduleId'),
            name=module.get('name'),
            cls=module.get('class'),
            rating=module.get('rating'),
            ship=module.get('ship'),
            modified=parse_ts(outfitting.get('updateTime'))
        )

def ingest_market(market):
    """Ingest station-level market data, yielding commodities."""
    if not market or not market.get('commodities'):
        return None
    for commodity in market['commodities']:
        yield Commodity(
            id=commodity.get('commodityId'),
            name=commodity.get('name', 'Unnamed'),
            category=commodity.get('category', 'Uncategorised'),
            demand=commodity.get('demand', 0),
            supply=commodity.get('supply', 0),
            sell=commodity.get('sellPrice', 0),
            buy=commodity.get('buyPrice', 0),
            modified=parse_ts(market.get('updateTime'))
        )

def parse_ts(ts):
    if ts is None:
        return None
    if ts.endswith('+00'):
        ts = ts[:-3]
    if '.' not in ts:
        ts += '.0'
    return datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f').replace(microsecond=0)



# Deprecated
from .parsing import MutuallyExclusiveGroup, ParseArgument
from ..tradeexcept import TradeException
from .exceptions import CommandLineError
from ..tradedb import System
from .. import prices, cache
import subprocess
import os
import sys
import pathlib

######################################################################
# Parser config

help = 'Update prices for a station.'
name = 'update'
epilog = ("Generates a human-readable version of the price list for a given station "
            "and opens it in the specified text editor.\n"
            "The format is intended to closely resemble the presentation of the "
            "market in-game. If you change the order items are listed in, "
            "the order will be kept for future edits, making it easier to quickly "
            "check for changes.")
wantsTradeDB = True
arguments = [
    ParseArgument('starting', help = 'Name of the station to update.', type = str)
]
switches = [
    ParseArgument('--timestamps', '-T',
            help = '[Text editing] Includes timestamps in the update.',
            action = 'store_true',
            default = False,
    ),
    ParseArgument('--all', '-A',
            help = 'List all known items.',
            action = 'store_true',
            default = False,
    ),
    ParseArgument('--use-demand', '-D',
            help = 'Unlock the "Demand" column in the GUI.',
            action = 'store_true',
            dest = 'useDemand',
            default = False,
    ),
    ParseArgument('--force-na', '-0',
            help = "Forces 'unk' supply to become 'n/a' by default",
            action = 'store_true',
            default = False,
            dest = 'forceNa',
    ),
    ParseArgument('--height', '-H',
            help = "[GUI] Specify height of the window",
            type = int,
            default = 800,
    ),
    ParseArgument('--front', '-F',
            help = (
                "[GUI] Keep the GUI infront of other windows; "
                "this allows you to put the window infront of "
                "the game UI if you run the game in windowed mode."
                ),
            action = 'store_true',
            default = False,
            dest = 'alwaysOnTop',
    ),
    ParseArgument('--window-x', '-wx',
            help = (
                "[GUI] Specify the window X position "
                "Use a negative value for a right-of-screen relative offset"
                ),
            default = -1,
            dest = 'windowX',
            type = int,
    ),
    ParseArgument('--window-y', '-wy',
            help = (
                "[GUI] Specify the window U position "
                "Use a negative value for a bottom-of-screen relative offset"
                ),
            default = 1,
            dest = 'windowY',
            type = int,
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--gui', '-G',
                help = "Use the experimental built-in GUI",
                action = 'store_true',
                default = False,
                dest = 'gui',
        ),
        ParseArgument('--editor',
                help = 'Generates a text file containing the prices for '
                        'the station and loads it into the specified '
                        'text editor for you.',
                default = None,
                type = str,
        ),
        ParseArgument('--sublime',
                help = 'Like --editor but uses Sublime Text (2 or 3).',
                action = 'store_const', const = 'sublime', dest = 'editing',
        ),
        ParseArgument('--notepad',
                help = 'Like --editor but uses Notepad.',
                action = 'store_const', const = 'notepad', dest = 'editing',
        ),
        ParseArgument('--npp',
                help = 'Like --editor but uses Notepad++.',
                action = 'store_const', const = 'npp', dest = 'editing',
        ),
        ParseArgument('--vim',
                help = 'Like --editor but uses vim.',
                action = 'store_const', const = 'vim', dest = 'editing',
        ),
    )
]

######################################################################
# Helpers


class TemporaryFileExistsError(TradeException):
    pass


def getTemporaryPath(cmdenv):
    tmpPath = pathlib.Path("prices.tmp")
    if tmpPath.exists():
        if not cmdenv.force:
            raise TemporaryFileExistsError(
                    "Temporary file already exists: {}\n"
                    "(Check you aren't already editing in another window"
                        .format(tmpPath)
            )
        tmpPath.unlink()
    return tmpPath


def saveTemporaryFile(tmpPath):
    if tmpPath.exists():
        lastPath = pathlib.Path("prices.last")
        if lastPath.exists():
            lastPath.unlink()
        tmpPath.rename(lastPath)


def saveCopyOfChanges(cmdenv, dbFilename, stationID):
    if "APPEND_PRICES" in os.environ:
        mode = "a"
    else:
        mode = "w"
    dumpPath = pathlib.Path("updated.prices")
    with dumpPath.open(mode, encoding = 'utf-8') as dumpFile:
        # Remember the filename so we know we need to delete it.
        prices.dumpPrices(dbFilename,
                prices.Element.full | prices.Element.blanks,
                file = dumpFile,
                stationID = stationID,
                defaultZero = False,
                debug = cmdenv.debug,
        )
    if not cmdenv.quiet:
        print("- Copy of changes saved as '{}'".format(
                str(dumpPath)
        ))


def getEditorPaths(cmdenv, editorName, envVar, windowsFolders, winExe, nixExe):
    cmdenv.DEBUG0("Locating {} editor", editorName)
    try:
        return os.environ[envVar]
    except KeyError:
        pass
    
    paths = []
    
    import platform
    system = platform.system()
    if system == 'Windows':
        binary = winExe
        for folder in ["Program Files", "Program Files (x86)"]:
            for version in windowsFolders:
                paths.append("{}\\{}\\{}".format(os.environ['SystemDrive'], folder, version))
    else:
        binary = nixExe
    
    try:
        paths += os.environ['PATH'].split(os.pathsep)
    except KeyError:
        pass
    
    for path in paths:
        candidate = os.path.join(path, binary)
        try:
            if pathlib.Path(candidate).exists():
                return candidate
        except OSError:
            pass
    
    raise CommandLineError(
        "ERROR: Unable to locate {} editor.\n"
        "Either specify the path to your editor with --editor "
        "or set the {} environment variable to point to it."
            .format(editorName, envVar)
    )


def editUpdate(tdb, cmdenv, stationID):
    """
        Dump the price data for a specific station to a file and
        launch the user's text editor to let them make changes
        to the file.
        
        If the user makes changes, re-load the file, update the
        database and regenerate the master .prices file.
    """
    
    cmdenv.DEBUG0("'update' mode with editor. editor:{} station:{}",
                    cmdenv.editor, cmdenv.origin)
    
    editor, editorArgs = cmdenv.editor, []
    if cmdenv.editing == 'sublime':
        cmdenv.DEBUG0("Sublime mode")
        editor = editor or \
                getEditorPaths(cmdenv,
                        "sublime", "SUBLIME_EDITOR",
                        ["Sublime Text 3", "Sublime Text 2"],
                        "sublime_text.exe",
                        "subl"
                )
        editorArgs += [ "--wait" ]
    elif cmdenv.editing == 'npp':
        cmdenv.DEBUG0("Notepad++ mode")
        editor = editor or \
                getEditorPaths(cmdenv,
                        "notepad++", "NOTEPADPP_EDITOR",
                        ["Notepad++"],
                        "notepad++.exe",
                        "notepad++"
                )
        if not cmdenv.quiet:
            print("NOTE: "
                    "You'll need to exit Notepad++ when you are done.")
    elif cmdenv.editing == "vim":
        cmdenv.DEBUG0("VI iMproved mode")
        if not editor:
            # Hack... Hackity hack, hack, hack.
            # This has a disadvantage in that: we don't check for just "vim" (no .exe) under Windows
            vimDirs = [
                    "Git\\share\\vim\\vim{}".format(vimVer)
                    for vimVer in range(70, 75)
            ]
            editor = getEditorPaths(cmdenv,
                    "vim",
                    "EDITOR",
                    vimDirs,
                    "vim.exe",
                    "vim"
            )
    elif cmdenv.editing == "notepad":
        cmdenv.DEBUG0("Notepad mode")
        editor = editor or "notepad.exe"  # herp
    
    try:
        envArgs = os.environ["EDITOR_ARGS"]
        if envArgs:
            editorArgs += envArgs.split(' ')
    except KeyError:
        pass
    
    # Create a temporary text file with a list of the price data.
    tmpPath = getTemporaryPath(cmdenv)
    
    absoluteFilename = None
    dbFilename = tdb.dbFilename
    try:
        elementMask = prices.Element.basic | prices.Element.supply
        if cmdenv.timestamps:
            elementMask |= prices.Element.timestamp
        if cmdenv.all:
            elementMask |= prices.Element.blanks
        # Open the file and dump data to it.
        with tmpPath.open("w", encoding = 'utf-8') as tmpFile:
            # Remember the filename so we know we need to delete it.
            absoluteFilename = str(tmpPath.resolve())
            prices.dumpPrices(dbFilename, elementMask,
                    file = tmpFile,
                    stationID = stationID,
                    defaultZero = cmdenv.forceNa,
                    debug = cmdenv.debug
            )
        
        # Stat the file so we can determine if the user writes to it.
        # Use the most recent create/modified timestamp.
        preStat = tmpPath.stat()
        preStamp = max(preStat.st_mtime, preStat.st_ctime)
        
        # Launch the editor
        editorCommandLine = [ editor ] + editorArgs + [ absoluteFilename ]
        cmdenv.DEBUG0("Invoking [{}]", ' '.join(editorCommandLine))
        try:
            result = subprocess.call(editorCommandLine)
        except FileNotFoundError:
            raise CommandLineError(
                        "Unable to launch requested editor: {}"
                            .format(editorCommandLine)
            )
        if result != 0:
            raise CommandLineError(
                    "NO DATA IMPORTED: "
                    "Your editor exited with a 'failed' exit code ({})"
                        .format(result)
            )
        
        # Did they update the file? Some editors destroy the file and rewrite it,
        # other files just write back to it, and some OSes do weird things with
        # these timestamps. That's why we have to use both mtime and ctime.
        postStat = tmpPath.stat()
        postStamp = max(postStat.st_mtime, postStat.st_ctime)
        
        if postStamp == preStamp:
            import random
            print("- No changes detected - doing nothing. {}".format(
                    random.choice([
                        "Brilliant!",
                        "I'll get my coat.",
                        "I ain't seen you.",
                        "You ain't seen me",
                        "... which was nice",
                        "Bingo!",
                        "Scorchio!",
                        "Boutros, boutros, ghali!",
                        "I'm Ed Winchester!",
                        "Suit you, sir! Oh!"
                    ])
            ))
        else:
            cache.importDataFromFile(tdb, cmdenv, tmpPath)
            saveCopyOfChanges(cmdenv, dbFilename, stationID)
        
        tmpPath.unlink()
        tmpPath = None
    
    except Exception as e:
        print("ERROR:", e)
        print()
        print("*** YOUR UPDATES WILL BE SAVED AS {} ***".format(
                "prices.last"
        ))
        # Save a copy
        if absoluteFilename and tmpPath:
            saveTemporaryFile(tmpPath)
        if "EXCEPTIONS" in os.environ:
            raise e


def guidedUpdate(tdb, cmdenv):
    dbFilename = tdb.dbFilename
    stationID = cmdenv.startStation.ID
    tmpPath = getTemporaryPath(cmdenv)
    
    cur = tdb.query("""
        SELECT  JULIANDAY('now') - JULIANDAY(MIN(modified)),
                JULIANDAY('now') - JULIANDAY(MAX(modified))
          FROM  StationItem
         WHERE  station_id = ?
    """, [stationID])
    oldest, newest = cur.fetchone()
    if oldest and newest:
        cmdenv.NOTE(
            "Current data {:.2f}-{:.2f} days old.",
            oldest, newest,
        )
    
    from .update_gui import render
    try:
        render(tdb, cmdenv, tmpPath)
        cmdenv.DEBUG0("Got results, importing")
        cache.importDataFromFile(tdb, cmdenv, tmpPath)
        
        saveCopyOfChanges(cmdenv, dbFilename, stationID)
        
        tmpPath.unlink()
        tmpPath = None
    
    except Exception as e:
        print("ERROR:", e)
        print()
        print("*** YOUR UPDATES WILL BE SAVED AS {} ***".format(
                "prices.last"
        ))
        
        if tmpPath:
            saveTemporaryFile(tmpPath)
        if "EXCEPTIONS" in os.environ:
            raise e

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    place = cmdenv.origPlace
    if isinstance(place, System):
        system = place
        if len(system.stations) != 1:
            raise CommandLineError(
                    "'{}' is a system, not a station.".format(
                        system.name()
            ))
        cmdenv.startStation = system.stations[0]
    else:
        cmdenv.startStation = place
    
    if cmdenv.gui or (not cmdenv.editor and not cmdenv.editing):
        if not cmdenv.quiet:
            print(
                    "NOTE:\n"
                    ". Press CTRL-C here to abort edit, otherwise "
                    "just close the update window to save.\n"
                    ". '-F' makes the window show in-front of the "
                    "E:D Window.\n"
                    ". '-A' forces all items to be listed.\n",
                    file = sys.stderr
            )
        guidedUpdate(tdb, cmdenv)
    else:
        # User specified one of the options to use an editor.
        editUpdate(tdb, cmdenv, cmdenv.startStation.ID)
    
    return None



"""
Import plugin that uses data files from 
https://elite.tromador.com/ to update the Database.
"""
from __future__ import annotations

from email.utils import parsedate_to_datetime
from pathlib import Path
from .. fs import file_line_count
from .. import plugins, cache, transfers
from ..misc import progress as pbar
from ..plugins import PluginException

import csv
import datetime
import os
import requests
import sqlite3
import typing


if typing.TYPE_CHECKING:
    from typing import Optional
    from .. tradeenv import TradeEnv

# Constants
BASE_URL = os.environ.get('TD_SERVER') or "https://elite.tromador.com/files/"


class DecodingError(PluginException):
    pass


def _count_listing_entries(tdenv: TradeEnv, listings: Path) -> int:
    """ Calculates the number of entries in a listing file by counting the lines. """
    if not listings.exists():
        tdenv.NOTE("File not found, aborting: {}", listings)
        return 0
    
    tdenv.DEBUG0(f"Getting total number of entries in {listings}...")
    count = file_line_count(listings)
    if count <= 1:
        if count == 1:
            tdenv.DEBUG0("Listing count of 1 suggests nothing but a header")
        else:
            tdenv.DEBUG0("Listings file is empty, nothing to do.")
        return 0
    
    return count + 1  # kfsone: Doesn't the header already make this + 1?


def _make_item_id_lookup(tdenv: TradeEnv, db: sqlite3.Cursor) -> frozenset[int]:
    """ helper: retrieve the list of commodities in database. """
    tdenv.DEBUG0("Getting list of commodities...")
    return frozenset(cols[0] for cols in db.execute("SELECT item_id FROM Item"))


def _make_station_id_lookup(tdenv: TradeEnv, db: sqlite3.Cursor) -> frozenset[int]:
    """ helper: retrieve the list of station IDs in database. """
    tdenv.DEBUG0("Getting list of stations...")
    return frozenset(cols[0] for cols in db.execute("SELECT station_id FROM Station"))


def _collect_station_modified_times(tdenv: TradeEnv, db: sqlite3.Cursor) -> dict[int, int]:
    """ helper: build a list of the last modified time for all stations by id. """
    tdenv.DEBUG0("Getting last-update times for stations...")
    return dict(db.execute("SELECT station_id, strftime('%s', MIN(modified)) FROM StationItem GROUP BY station_id"))


class ImportPlugin(plugins.ImportPluginBase):
    """
    Plugin that downloads data from eddb.
    """
    pluginOptions = {
        'item':         "Update Items using latest file from server. (Implies '-O system,station')",
        'rare':         "Update RareItems using latest file from server. (Implies '-O system,station')",
        'ship':         "Update Ships using latest file from server.",
        'upgrade':      "Update Upgrades using latest file from server.",
        'system':       "Update Systems using latest file from server.",
        'station':      "Update Stations using latest file from server. (Implies '-O system')",
        'shipvend':     "Update ShipVendors using latest file from server. (Implies '-O system,station,ship')",
        'upvend':       "Update UpgradeVendors using latest file from server. (Implies '-O system,station,upgrade')",
        'listings':     "Update market data using latest listings.csv dump. (Implies '-O item,system,station')",
        'all':          "Update everything with latest dumpfiles. (Regenerates all tables)",
        'clean':        "Erase entire database and rebuild from empty. (Regenerates all tables.)",
        'skipvend':     "Don't regenerate ShipVendors or UpgradeVendors. (Supercedes '-O all', '-O clean'.)",
        'force':        "Force regeneration of selected items even if source file not updated since previous run. "
                        "(Useful for updating Vendor tables if they were skipped during a '-O clean' run.)",
        'purge':        "Remove any empty systems that previously had fleet carriers.",
        'optimize':     "Optimize ('vacuum') database after processing.",
        'solo':         "Don't download crowd-sourced market data. (Implies '-O skipvend', supercedes '-O all', '-O clean', '-O listings'.)",
    }
    
    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)
        
        self.dataPath = os.environ.get('TD_EDDB') or self.tdenv.tmpDir
        self.categoriesPath = Path("Category.csv")
        self.commoditiesPath = Path("Item.csv")
        self.rareItemPath = Path("RareItem.csv")
        self.shipPath = Path("Ship.csv")
        self.urlShipyard = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/shipyard.csv"
        self.FDevShipyardPath = self.tdb.dataPath / Path("FDevShipyard.csv")
        self.shipVendorPath = Path("ShipVendor.csv")
        self.stationsPath = Path("Station.csv")
        self.sysPath = Path("System.csv")
        self.upgradesPath = Path("Upgrade.csv")
        self.urlOutfitting = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/outfitting.csv"
        self.FDevOutfittingPath = self.tdb.dataPath / Path("FDevOutfitting.csv")
        self.upgradeVendorPath = Path("UpgradeVendor.csv")
        self.listingsPath = Path("listings.csv")
        self.liveListingsPath = Path("listings-live.csv")
        self.pricesPath = Path("listings.prices")
    
    def now(self):
        return datetime.datetime.now()
    
    def downloadFile(self, path):
        """
        Fetch the latest dumpfile from the website if newer than local copy.
        """
        if path not in (self.liveListingsPath, self.listingsPath):
            localPath = Path(self.tdb.dataPath, path)
        else:
            localPath = Path(self.dataPath, path)
        
        url  = BASE_URL + str(path)
        
        self.tdenv.NOTE("Checking for update to '{}'.", path)
        # Use an HTTP Request header to obtain the Last-Modified and Content-Length headers.
        # Also, tell the server to give us the un-compressed length of the file by saying
        # that >this< request only wants text.
        headers = {"User-Agent": "Trade-Dangerous", "Accept-Encoding": "identity"}
        try:
            response = requests.head(url, headers=headers, timeout=70)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.tdenv.WARN("Problem with download:\n    URL: {}\n    Error: {}", url, str(e))
            return False
        
        last_modified = response.headers.get("last-modified")
        dump_mod_time = parsedate_to_datetime(last_modified).timestamp()
        
        if Path.exists(localPath):
            local_mod_time = localPath.stat().st_mtime
            if local_mod_time >= dump_mod_time:
                self.tdenv.DEBUG0("'{}': Dump is not more recent than Local.", path)
                return False
        
        # The server doesn't know the gzip'd length, and we won't see the gzip'd data,
        # so we want the actual text-only length. Capture it here so we can tell the
        # transfer mechanism how big the file is going to be.
        length = response.headers.get("content-length")
        
        self.tdenv.NOTE("Downloading file '{}'.", path)
        transfers.download(self.tdenv, url, localPath, chunkSize=16384, length=length)
        
        # Change the timestamps on the file so they match the website
        os.utime(localPath, (dump_mod_time, dump_mod_time))
        
        return True
    
    def purgeSystems(self):
        """
        Purges systems from the System table that do not have any stations claiming to be in them.
        Keeps table from becoming too large because of fleet carriers moving to unpopulated systems.
        """
        db = self.tdb.getDB()
        self.tdenv.NOTE("Purging Systems with no stations: Start time = {}", self.now())
        
        db.execute("""
            DELETE FROM System
             WHERE NOT EXISTS(SELECT 1 FROM Station WHERE Station.system_id = System.system_id)
        """)
        db.commit()
        
        self.tdenv.NOTE("Finished purging Systems. End time = {}", self.now())
    
    def importListings(self, listings_file):
        """
        Updates the market data (AKA the StationItem table) using listings_file
        Writes directly to database.
        """
        listings_path = Path(self.dataPath, listings_file).absolute()
        from_live = listings_path != Path(self.dataPath, self.listingsPath).absolute()
        
        self.tdenv.NOTE("Checking listings")
        total = _count_listing_entries(self.tdenv, listings_path)
        if not total:
            self.tdenv.NOTE("No listings")
            return
        
        self.tdenv.NOTE("Processing market data from {}: Start time = {}. Live = {}", listings_file, self.now(), from_live)
        
        db = self.tdb.getDB()
        stmt_unliven_station = """UPDATE StationItem SET from_live = 0 WHERE station_id = ?"""
        stmt_flush_station   = """DELETE from StationItem WHERE station_id = ?"""
        stmt_add_listing     = """
            INSERT OR IGNORE INTO StationItem (
                station_id, item_id, modified, from_live,
                demand_price, demand_units, demand_level,
                supply_price, supply_units, supply_level
            )
            VALUES (
                ?, ?, datetime(?, 'unixepoch'), ?,
                ?, ?, ?,
                ?, ?, ?
            )
        """
        
        # Fetch all the items IDS
        item_lookup = _make_item_id_lookup(self.tdenv, db.cursor())
        station_lookup = _make_station_id_lookup(self.tdenv, db.cursor())
        last_station_update_times = _collect_station_modified_times(self.tdenv, db.cursor())
        
        cur_station = None
        is_debug = self.tdenv.debug > 0
        self.tdenv.DEBUG0("Processing entries...")
        
        # Try to find a balance between doing too many commits where we fail
        # to get any benefits from constructing transactions, and blowing up
        # the WAL and memory usage by making massive transactions.
        max_transaction_items, transaction_items = 32 * 1024, 0
        with pbar.Progress(total, 40, prefix="Processing", style=pbar.LongRunningCountBar) as prog,\
              listings_path.open("r", encoding="utf-8", errors="ignore") as fh:
            cursor = db.cursor()
            cursor.execute("BEGIN TRANSACTION")
            
            for listing in csv.DictReader(fh):
                prog.increment(1)
                
                station_id = int(listing['station_id'])
                if station_id not in station_lookup:
                    continue
                
                listing_time = int(listing['collected_at'])
                
                if station_id != cur_station:
                    # commit anything from the previous station, get a new cursor
                    if transaction_items >= max_transaction_items:
                        cursor.execute("COMMIT")
                        transaction_items = 0
                        cursor.execute("BEGIN TRANSACTION")
                    cur_station, skip_station = station_id, False
                    
                    # Check if listing already exists in DB and needs updated.
                    last_modified: int = int(last_station_update_times.get(station_id, 0))
                    if last_modified:
                        # When the listings.csv data matches the database, update to make from_live == 0.
                        if listing_time == last_modified and not from_live:
                            if is_debug:
                                self.tdenv.DEBUG1(f"Marking {cur_station} as no longer 'live' (old={last_modified}, listing={listing_time}).")
                            cursor.execute(stmt_unliven_station, (cur_station,))
                            transaction_items += 1
                            skip_station = True
                            continue
                        
                        # Unless the import file data is newer, nothing else needs to be done for this station,
                        # so the rest of the listings for this station can be skipped.
                        if listing_time <= last_modified:
                            skip_station = True
                            continue
                        
                        # The data from the import file is newer, so we need to delete the old data for this station.
                        if is_debug:
                            self.tdenv.DEBUG1(f"Deleting old listing data for {cur_station} (old={last_modified}, listing={listing_time}).")
                        cursor.execute(stmt_flush_station, (cur_station,))
                        transaction_items += 1
                        last_station_update_times[station_id] = listing_time
                
                # station skip lasts until we change station id.
                if skip_station:
                    continue
                
                # Since this station is not being skipped, get the data and prepare for insertion into the DB.
                item_id = int(listing['commodity_id'])
                # listings.csv includes rare items, which we are ignoring.
                if item_id not in item_lookup:
                    continue
                
                demand_price = int(listing['sell_price'])
                demand_units = int(listing['demand'])
                demand_level = int(listing.get('demand_bracket') or '-1')
                supply_price = int(listing['buy_price'])
                supply_units = int(listing['supply'])
                supply_level = int(listing.get('supply_bracket') or '-1')
                
                if is_debug:
                    self.tdenv.DEBUG1(f"Inserting new listing data for {station_id}.")
                cursor.execute(stmt_add_listing, (
                        station_id, item_id, listing_time, from_live,
                        demand_price, demand_units, demand_level,
                        supply_price, supply_units, supply_level,
                ))
                transaction_items += 1
        
        # These will take a little while, which has four steps, so we'll make it a counter.
        with pbar.Progress(1, 40, prefix="Saving"):
            # Do a final commit to be sure
            cursor.execute("COMMIT")
        
        if self.getOption("optimize"):
            with pbar.Progress(1, 40, prefix="Optimizing"):
                db.execute("VACUUM")
        
        self.tdb.close()
        
        self.tdenv.NOTE("Finished processing market data. End time = {}", self.now())
    
    def run(self):
        self.tdenv.ignoreUnknown = True
        
        # Create the /eddb folder for downloading the source files if it doesn't exist.
        try:
            Path(str(self.dataPath)).mkdir()
        except FileExistsError:
            pass
        
        # Run 'listings' by default:
        # If no options, or if only 'force', and/or 'skipvend',
        # have been passed, enable 'listings'.
        default = True
        for option in self.options:
            if option not in ('force', 'skipvend', 'purge'):
                default = False
        if default:
            self.options["listings"] = True
        
        # We can probably safely assume that the plugin
        # has never been run if the db file doesn't exist.
        if not self.tdb.dbPath.exists():
            self.options["clean"] = True
        
        if self.getOption("clean"):
            # Rebuild the tables from scratch. Must be done on first run of plugin.
            # Can be done at anytime with the "clean" option.
            for name in [
                "Category",
                "Item",
                "RareItem",
                "Ship",
                "ShipVendor",
                "Station",
                "System",
                "Upgrade",
                "UpgradeVendor",
                "FDevShipyard",
                "FDevOutfitting",
            ]:
                file = self.tdb.dataPath / Path(name + ".csv")
                try:
                    os.remove(str(file))
                except FileNotFoundError:
                    pass
            
            try:
                os.remove(str(self.tdb.dataPath) + "/TradeDangerous.db")
            except FileNotFoundError:
                pass
            try:
                os.remove(str(self.tdb.dataPath) + "/TradeDangerous.prices")
            except FileNotFoundError:
                pass
            
            # Because this is a clean run, we need to temporarily rename the RareItem.csv,
            # otherwise TD will crash trying to insert the rare items to the database,
            # because there's nothing in the Station table it tries to pull from.
            ri_path = self.tdb.dataPath / Path("RareItem.csv")
            rib_path = ri_path.with_suffix(".tmp")
            if ri_path.exists():
                if rib_path.exists():
                    rib_path.unlink()
                ri_path.rename(rib_path)
            
            self.tdb.close()
            
            self.tdb.reloadCache()
            self.tdb.close()
            
            # Now it's safe to move RareItems back.
            if ri_path.exists():
                ri_path.unlink()
            if rib_path.exists():
                rib_path.rename(ri_path)
            
            self.options["all"] = True
            self.options["force"] = True
        
        # Select which options will be updated
        if self.getOption("listings"):
            self.options["item"] = True
            self.options["station"] = True
        
        if self.getOption("shipvend"):
            self.options["ship"] = True
            self.options["station"] = True
        
        if self.getOption("upvend"):
            self.options["upgrade"] = True
            self.options["station"] = True
        
        if self.getOption("item"):
            self.options["station"] = True
        
        if self.getOption("rare"):
            self.options["station"] = True
        
        if self.getOption("station"):
            self.options["system"] = True
        
        if self.getOption("all"):
            self.options["item"] = True
            self.options["rare"] = True
            self.options["ship"] = True
            self.options["shipvend"] = True
            self.options["station"] = True
            self.options["system"] = True
            self.options["upgrade"] = True
            self.options["upvend"] = True
            self.options["listings"] = True
        
        if self.getOption("solo"):
            self.options["listings"] = False
            self.options["skipvend"] = True
        
        if self.getOption("skipvend"):
            self.options["shipvend"] = False
            self.options["upvend"] = False
        
        # Download required files and update tables.
        buildCache = False
        if self.getOption("upgrade"):
            if self.downloadFile(self.upgradesPath) or self.getOption("force"):
                transfers.download(self.tdenv, self.urlOutfitting, self.FDevOutfittingPath)
                buildCache = True
        
        if self.getOption("ship"):
            if self.downloadFile(self.shipPath) or self.getOption("force"):
                transfers.download(self.tdenv, self.urlShipyard, self.FDevShipyardPath)
                buildCache = True
        
        if self.getOption("rare"):
            if self.downloadFile(self.rareItemPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("shipvend"):
            if self.downloadFile(self.shipVendorPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("upvend"):
            if self.downloadFile(self.upgradeVendorPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("system"):
            if self.downloadFile(self.sysPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("station"):
            if self.downloadFile(self.stationsPath) or self.getOption("force"):
                buildCache = True
        
        if self.getOption("item"):
            if self.downloadFile(self.commoditiesPath) or self.getOption("force"):
                self.downloadFile(self.categoriesPath)
                buildCache = True
        
        # Remake the .db files with the updated info.
        if buildCache:
            self.tdb.close()
            self.tdb.reloadCache()
            self.tdb.close()
        
        if self.getOption("purge"):
            self.purgeSystems()
            self.tdb.close()
        
        if self.getOption("listings"):
            if self.downloadFile(self.listingsPath) or self.getOption("force"):
                self.importListings(self.listingsPath)
            if self.downloadFile(self.liveListingsPath) or self.getOption("force"):
                self.importListings(self.liveListingsPath)
        
        if self.getOption("listings"):
            self.tdenv.NOTE("Regenerating .prices file.")
            cache.regeneratePricesFile(self.tdb, self.tdenv)
        
        self.tdenv.NOTE("Import completed.")
        
        # TD doesn't need to do anything, tell it to just quit.
        return False



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018, 2019
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Cache loader
#
#  TD works primarily from an SQLite3 database, but the data in that
#  is sourced from text files.
#   data/TradeDangerous.sql contains the less volatile data - systems,
#   ships, etc
#   data/TradeDangerous.prices contains a description of the price
#   database that is intended to be easily editable and commitable to
#   a source repository.
#
#  TODO: Split prices into per-system or per-station files so that
#  we can tell how old data for a specific system is.

from __future__ import annotations

from pathlib import Path
import csv
import os
import re
import sqlite3
import sys
import typing

from functools import partial as partial_fn
from .fs import file_line_count
from .tradeexcept import TradeException
from tradedangerous.misc.progress import Progress, CountingBar
from . import corrections, utils
from . import prices


# For mypy/pylint type checking
if typing.TYPE_CHECKING:
    from typing import Any, Callable, Optional, TextIO  # noqa
    
    from .tradeenv import TradeEnv


######################################################################
# Regular expression patterns. Here be draegons.
# If you add new patterns:
# - use fragments and re.VERBOSE (see itemPriceRe)
# - use named captures (?P<name> ...)
# - include comments

# # Match the '@ SYSTEM/Station' line
systemStationRe = re.compile(r'^\@\s*(.*)/(.*)')

# # Price Line matching

# first part of any prices line is the item name and paying/asking price
itemPriceFrag = r"""
    # match item name, allowing spaces in the name
    (?P<item> .*?)
\s+
    # price station is buying the item for
    (?P<sell> \d+)
\s+
    # price station is selling item for
    (?P<buy> \d+)
"""

# time formats per https://www.sqlite.org/lang_datefunc.html
# YYYY-MM-DD HH:MM:SS
# YYYY-MM-DDTHH:MM:SS
# HH:MM:SS
# 'now'
timeFrag = r'(?P<time>(\d{4}-\d{2}-\d{2}[T ])?\d{2}:\d{2}:\d{2}|now)'

# <name> <sell> <buy> [ <demand> <supply> [ <time> | now ] ]
qtyLevelFrag = r"""
    unk             # You can just write 'unknown'
|   \?              # alias for unknown
|   n/a             # alias for 0L0
|   -               # alias for 0L0
|   \d+[\?LMH]      # Or <number><level> where level is L(ow), M(ed) or H(igh)
|   0               # alias for n/a
|   bug
"""
newItemPriceRe = re.compile(r"""
^
    {base_f}
    (
    \s+
        # demand units and level
        (?P<demand> {qtylvl_f})
    \s+
        # supply units and level
        (?P<supply> {qtylvl_f})
        # time is optional
        (?:
        \s+
            {time_f}
        )?
    )?
\s*
$
""".format(base_f = itemPriceFrag, qtylvl_f = qtyLevelFrag, time_f = timeFrag),
            re.IGNORECASE + re.VERBOSE)

######################################################################
# Exception classes


class BuildCacheBaseException(TradeException):
    """
    Baseclass for BuildCache exceptions
    Attributes:
        fileName    Name of file being processedStations
        lineNo      Line the error occurred on
        error       Description of the error
    """
    
    def __init__(self, fromFile: Path, lineNo: int, error: str = None) -> None:
        self.fileName = fromFile.name
        self.lineNo = lineNo
        self.category = "ERROR"
        self.error = error or "UNKNOWN ERROR"
    
    def __str__(self) -> str:
        return f'{self.fileName}:{self.lineNo} {self.category} {self.error}'


class UnknownSystemError(BuildCacheBaseException):
    """
    Raised when the file contains an unknown star name.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, key: str) -> None:
        super().__init__(fromFile, lineNo, f'Unrecognized SYSTEM: "{key}"')


class UnknownStationError(BuildCacheBaseException):
    """
    Raised when the file contains an unknown star/station name.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, key: str) -> None:
        super().__init__(fromFile, lineNo, f'Unrecognized STAR/Station: "{key}"')


class UnknownItemError(BuildCacheBaseException):
    """
    Raised in the case of an item name that we don't know.
    Attributes:
        itemName   Key we tried to look up.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, itemName: str) -> None:
        super().__init__(fromFile, lineNo, f'Unrecognized item name: "{itemName}"')


class DuplicateKeyError(BuildCacheBaseException):
    """
        Raised when an item is being redefined.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, keyType: str, keyValue: str, prevLineNo: int) -> None:
        super().__init__(fromFile, lineNo,
                         f'Second occurrance of {keyType} "{keyValue}", previous entry at line {prevLineNo}.')


class DeletedKeyError(BuildCacheBaseException):
    """
    Raised when a key value in a .csv file is marked as DELETED in the
    corrections file.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, keyType: str, keyValue: str) -> None:
        super().__init__(
            fromFile, lineNo,
            f'{keyType} "{keyValue}" is marked as DELETED and should not be used.'
        )


class DeprecatedKeyError(BuildCacheBaseException):
    """
    Raised when a key value in a .csv file has a correction; the old
    name should not appear in the .csv file.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, keyType: str, keyValue: str, newValue: str) -> None:
        super().__init__(
            fromFile, lineNo,
            f'{keyType} "{keyValue}" is deprecated and should be replaced with "{newValue}".'
        )


class MultipleStationEntriesError(DuplicateKeyError):
    """ Raised when a station appears multiple times in the same file. """
    
    def __init__(self, fromFile: Path, lineNo: int, facility: str, prevLineNo: int) -> None:
        super().__init__(fromFile, lineNo, 'station', facility, prevLineNo)


class MultipleItemEntriesError(DuplicateKeyError):
    """ Raised when one item appears multiple times in the same station. """
    
    def __init__(self, fromFile: Path, lineNo: int, item: str, prevLineNo: int) -> None:
        super().__init__(fromFile, lineNo, 'item', item, prevLineNo)


class InvalidLineError(BuildCacheBaseException):
    """
    Raised when an invalid line is read.
    Attributes:
        problem     The problem that occurred
        text        Offending text
    """
    
    def __init__(self, fromFile: Path, lineNo: int, problem: str, text: str) -> None:
        super().__init__(fromFile, lineNo, f'{problem},\ngot: "{text.strip()}".')


class SupplyError(BuildCacheBaseException):
    """
    Raised when a supply field is incorrectly formatted.
    """
    
    def __init__(self, fromFile: Path, lineNo: int, category: str, problem: str, value: Any) -> None:
        super().__init__(fromFile, lineNo, f'Invalid {category} supply value: {problem}. Got: {value}')


######################################################################
# Helpers


# supply/demand levels are one of '?' for unknown, 'L', 'M' or 'H'
# for low, medium, or high. We turn these into integer values for
# ordering convenience, and we include both upper and lower-case
# so we don't have to sweat ordering.
#
SUPPLY_LEVEL_VALUES = {
    '?':   -1,
    'L':    1,      'l':    1,
    'M':    2,      'm':    2,
    'H':    3,      'h':    3,
}


def parseSupply(pricesFile: Path, lineNo: int, category: str, reading: str) -> tuple[int, int]:
    """ Parse a supply specifier which is expected to be in the <number><?, L, M, or H>, and
        returns the units as an integer and a numeric level value suitable for ordering,
        such that ? = -1, L/l = 0, M/m = 1, H/h = 2 """
    
    #   supply_level <- digit+ level;
    #   digit <- [0-9];
    #   level <- Unknown / Low / Medium / High;
    #   Unknown <- '?';
    #   Low <- 'L';
    #   Medium <- 'M';
    #   High <- 'H';
    if reading == '?':
        return -1, -1
    if reading == '-':
        return 0, 0
    
    # extract the left most digits into unit and the last character into the level reading.
    units, level = reading[0:-1], reading[-1]
    
    # Extract the right most character as the "level" and look up its numeric value.
    levelNo = SUPPLY_LEVEL_VALUES.get(level)
    if levelNo is None:
        raise SupplyError(
            pricesFile, lineNo, category, reading,
            f'Unrecognized level suffix: "{level}": expected one of "L", "M", "H" or "?"'
        )
    
    # Expecting a numeric value in units, e.g. 123? -> (units=123, level=?)
    try:
        unitsNo = int(units)
        if unitsNo < 0:
            # Use the same code-path as if the units fail to parse.
            raise ValueError('negative unit count')
    except ValueError:
        raise SupplyError(
            pricesFile, lineNo, category, reading,
            f'Unrecognized units/level value: "{level}": expected "-", "?", or a number followed by a level (L, M, H or ?).'
        ) from None  # don't forward the exception itself
    
    # Normalize the units and level when there are no units.
    if unitsNo == 0:
        return 0, 0
    
    return unitsNo, levelNo


######################################################################
# Code
######################################################################


def getSystemByNameIndex(cur: sqlite3.Cursor) -> dict[str, int]:
    """ Build station index in STAR/Station notation """
    cur.execute("""
            SELECT system_id, UPPER(system.name)
              FROM System
        """)
    return { name: ID for (ID, name) in cur }


def getStationByNameIndex(cur: sqlite3.Cursor) -> dict[str, int]:
    """ Build station index in STAR/Station notation """
    cur.execute("""
            SELECT station_id,
                    system.name || '/' || station.name
              FROM System
                   INNER JOIN Station
                      USING (system_id)
        """)
    return { name.upper(): ID for (ID, name) in cur }


def getItemByNameIndex(cur: sqlite3.Cursor) -> dict[str, int]:
    """
        Generate item name index.
    """
    cur.execute("SELECT item_id, UPPER(name) FROM item")
    return { name: itemID for (itemID, name) in cur }


# The return type of process prices is complicated, should probably have been a type
# in its own right. I'm going to define some aliases to try and persuade IDEs to be
# more helpful about what it is trying to return.
if typing.TYPE_CHECKING:
    # A list of the IDs of stations that were modified so they can be updated
    ProcessedStationIds= tuple[tuple[int]]
    ProcessedItem = tuple[
        int,                            # station ID
        int,                            # item ID
        Optional[int | float |str],     # modified
        int,                            # demandCR
        int,                            # demandUnits
        int,                            # demandLevel
        int,                            # supplyCr
        int,                            # supplyUnits
        int,                            # supplyLevel
    ]
    ProcessedItems = list[ProcessedItem]
    ZeroItems = list[tuple[int, int]]   # stationID, itemID


def processPrices(tdenv: TradeEnv, priceFile: Path, db: sqlite3.Connection, defaultZero: bool) -> tuple[ProcessedStationIds, ProcessedItems, ZeroItems, int, int, int, int]:
    """
        Yields SQL for populating the database with prices
        by reading the file handle for price lines.
        
        :param tdenv:       The environment we're working in
        :param priceFile:   File to read
        :param db:          SQLite3 database to write to
        :param defaultZero: Whether to create default zero-availability/-demand records for data that's not present
                            (if this is a partial update, you don't want this to be False)
    """
    
    DEBUG0, DEBUG1 = tdenv.DEBUG0, tdenv.DEBUG1
    DEBUG0("Processing prices file: {}", priceFile)
    
    cur = db.cursor()
    ignoreUnknown = tdenv.ignoreUnknown
    quiet = tdenv.quiet
    merging = tdenv.mergeImport
    
    systemByName = getSystemByNameIndex(cur)
    stationByName = getStationByNameIndex(cur)
    stationByName.update(
        (sys, ID)
        for sys, ID in corrections.stations.items()
        if isinstance(ID, int)
    )
    sysCorrections = corrections.systems
    stnCorrections = {
        stn: alt
        for stn, alt in corrections.stations.items()
        if isinstance(alt, str)
    }
    
    itemByName = getItemByNameIndex(cur)
    
    defaultUnits = -1 if not defaultZero else 0
    defaultLevel = -1 if not defaultZero else 0
    
    stationID = None
    facility = None
    processedStations = {}
    processedSystems = set()
    processedItems = {}
    stationItemDates = {}
    DELETED = corrections.DELETED
    items, zeros = [], []
    
    lineNo, localAdd = 0, 0
    if not ignoreUnknown:
        def ignoreOrWarn(error: Exception) -> None:
            raise error
    
    elif not quiet:
        ignoreOrWarn = tdenv.WARN
    
    def changeStation(matches: re.Match) -> None:
        nonlocal facility, stationID
        nonlocal processedStations, processedItems, localAdd
        nonlocal stationItemDates
        
        # ## Change current station
        stationItemDates = {}
        systemNameIn, stationNameIn = matches.group(1, 2)
        systemName, stationName = systemNameIn.upper(), stationNameIn.upper()
        corrected = False
        facility = f'{systemName}/{stationName}'
        
        # Make sure it's valid.
        stationID = DELETED
        newID = stationByName.get(facility, -1)  # why -1 and not None?
        DEBUG0("Selected station: {}, ID={}", facility, newID)
        if newID is DELETED:
            DEBUG1("DELETED Station: {}", facility)
            return
        
        if newID < 0:
            if utils.checkForOcrDerp(tdenv, systemName, stationName):
                return
            corrected = True
            altName = sysCorrections.get(systemName)
            if altName is DELETED:
                DEBUG1("DELETED System: {}", facility)
                return
            if altName:
                DEBUG1("SYSTEM '{}' renamed '{}'", systemName, altName)
                systemName, facility = altName, "/".join((altName, stationName))
            
            systemID = systemByName.get(systemName, -1)  # why -1 and not None?
            if systemID < 0:
                ignoreOrWarn(
                    UnknownSystemError(priceFile, lineNo, facility)
                )
                return
            
            altStation = stnCorrections.get(facility)
            if altStation:
                if altStation is DELETED:
                    DEBUG1("DELETED Station: {}", facility)
                    return
                
                DEBUG1("Station '{}' renamed '{}'", facility, altStation)
                stationName = altStation.upper()
                facility = f'{systemName}/{stationName}'
            
            newID = stationByName.get(facility, -1)
            if newID is DELETED:
                DEBUG1("Renamed station DELETED: {}", facility)
                return
        
        if newID < 0:
            if not ignoreUnknown:
                DEBUG0(f'Key value: "{list(stationByName.keys())[list(stationByName.values()).index(128893178)]}"')
                ignoreOrWarn(
                    UnknownStationError(priceFile, lineNo, facility)
                )
                return
            name = utils.titleFixup(stationName)
            inscur = db.cursor()
            inscur.execute("""
                INSERT INTO Station (
                    system_id, name,
                    ls_from_star,
                    blackmarket,
                    max_pad_size,
                    market,
                    shipyard,
                    modified
                ) VALUES (
                    ?, ?, 0, '?', '?', '?', '?',
                    DATETIME('now')
                )
            """, [systemID, name])
            newID = inscur.lastrowid
            stationByName[facility] = newID
            tdenv.NOTE(
                "Added local station placeholder for {} (#{})", facility, newID
            )
            localAdd += 1
        elif newID in processedStations:
            # Check for duplicates
            if not corrected:
                raise MultipleStationEntriesError(
                    priceFile, lineNo, facility,
                    processedStations[newID]
                )
        
        stationID = newID
        processedSystems.add(systemName)
        processedStations[stationID] = lineNo
        processedItems = {}
        
        cur = db.execute("""
            SELECT item_id, modified
              FROM StationItem
             WHERE station_id = ?
        """, [stationID])
        stationItemDates = dict(cur)
    
    addItem, addZero = items.append, zeros.append
    getItemID = itemByName.get
    newItems, updtItems, ignItems = 0, 0, 0
    
    def processItemLine(matches):
        nonlocal newItems, updtItems, ignItems
        itemName, modified = matches.group('item', 'time')
        itemName = itemName.upper()
        
        # Look up the item ID.
        itemID = getItemID(itemName, -1)
        if itemID < 0:
            oldName = itemName
            itemName = corrections.correctItem(itemName)
            if itemName == DELETED:
                DEBUG1("DELETED {}", oldName)
                return
            itemName = itemName.upper()
            itemID = getItemID(itemName, -1)
            if itemID < 0:
                ignoreOrWarn(
                    UnknownItemError(priceFile, lineNo, itemName)
                )
                return
            DEBUG1("Renamed {} -> {}", oldName, itemName)
        
        lastModified = stationItemDates.get(itemID, None)
        if lastModified and merging:
            if modified and modified != 'now' and modified <= lastModified:
                DEBUG1("Ignoring {} @ {}: {} <= {}".format(
                    itemName, facility,
                    modified, lastModified,
                ))
                if modified < lastModified:
                    ignItems += 1
                return
        
        # Check for duplicate items within the station.
        if itemID in processedItems:
            ignoreOrWarn(
                MultipleItemEntriesError(
                    priceFile, lineNo,
                    f'{itemName}',
                    processedItems[itemID]
                )
            )
            return
        
        demandCr, supplyCr = matches.group('sell', 'buy')
        demandCr, supplyCr = int(demandCr), int(supplyCr)
        demandString, supplyString = matches.group('demand', 'supply')
        
        if demandCr == 0 and supplyCr == 0:
            if lastModified:
                addZero((stationID, itemID))
        else:
            if lastModified:
                updtItems += 1
            else:
                newItems += 1
            if demandString:
                demandUnits, demandLevel = parseSupply(
                    priceFile, lineNo, 'demand', demandString
                )
            else:
                demandUnits, demandLevel = defaultUnits, defaultLevel
            
            if demandString and supplyString:
                supplyUnits, supplyLevel = parseSupply(
                    priceFile, lineNo, 'supply', supplyString
                )
            else:
                supplyUnits, supplyLevel = defaultUnits, defaultLevel
            
            if modified == 'now':
                modified = None  # Use CURRENT_FILESTAMP
            
            addItem((
                stationID, itemID, modified,
                demandCr, demandUnits, demandLevel,
                supplyCr, supplyUnits, supplyLevel,
            ))
        
        processedItems[itemID] = lineNo
    
    space_cleanup = re.compile(r'\s{2,}').sub
    for line in priceFile:
        lineNo += 1
        
        text = line.split('#', 1)[0]                # Discard comments
        text = space_cleanup(' ', text).strip()     # Remove leading/trailing whitespace, reduce multi-spaces
        if not text:
            continue
        
        ########################################
        # ## "@ STAR/Station" lines.
        if text.startswith('@'):
            matches = systemStationRe.match(text)
            if not matches:
                raise InvalidLineError(priceFile, lineNo, "Unrecognized '@' line", text)
            changeStation(matches)
            continue
        
        if not stationID:
            # Need a station to process any other type of line.
            raise InvalidLineError(priceFile, lineNo, "Expecting '@ SYSTEM / Station' line", text)
        if stationID == DELETED:
            # Ignore all values from a deleted station/system.
            continue
        
        ########################################
        # ## "+ Category" lines
        if text.startswith('+'):
            # we now ignore these.
            continue
        
        ########################################
        # ## "Item sell buy ..." lines.
        matches = newItemPriceRe.match(text)
        if not matches:
            raise InvalidLineError(priceFile, lineNo, "Unrecognized line/syntax", text)
        
        processItemLine(matches)
    
    numSys = len(processedSystems)
    
    if localAdd > 0:
        tdenv.NOTE(
            "Placeholder stations are added to the local DB only "
            "(not the .CSV).\n"
            "Use 'trade.py export --table Station' "
            "if you /need/ to persist them."
        )
    
    stations = tuple((ID,) for ID in processedStations)
    return stations, items, zeros, newItems, updtItems, ignItems, numSys


######################################################################


def processPricesFile(tdenv: TradeEnv, db: sqlite3.Connection, pricesPath: Path, pricesFh: Optional[TextIO] = None, defaultZero: bool = False) -> None:
    tdenv.DEBUG0("Processing Prices file '{}'", pricesPath)
    
    with (pricesFh or pricesPath.open('r', encoding='utf-8')) as fh:
        stations, items, zeros, newItems, updtItems, ignItems, numSys = processPrices(
            tdenv, fh, db, defaultZero
        )
    
    if not tdenv.mergeImport:
        db.executemany("""
            DELETE FROM StationItem
             WHERE station_id = ?
        """, stations)
    if zeros:
        db.executemany("""
            DELETE FROM StationItem
             WHERE station_id = ?
               AND item_id = ?
        """, zeros)
    removedItems = len(zeros)
    
    if items:
        for item in items:
            try:
                db.execute("""
                    INSERT OR REPLACE INTO StationItem (
                        station_id, item_id, modified,
                        demand_price, demand_units, demand_level,
                        supply_price, supply_units, supply_level
                    ) VALUES (
                        ?, ?, IFNULL(?, CURRENT_TIMESTAMP),
                        ?, ?, ?,
                        ?, ?, ?
                    )
                """, item)
            except sqlite3.IntegrityError as e:
                print(e)
                print(item)
                raise e
        # db.executemany("""
        #     INSERT OR REPLACE INTO StationItem (
        #         station_id, item_id, modified,
        #         demand_price, demand_units, demand_level,
        #         supply_price, supply_units, supply_level
        #     ) VALUES (
        #         ?, ?, IFNULL(?, CURRENT_TIMESTAMP),
        #         ?, ?, ?,
        #         ?, ?, ?
        #     )
        # """, items)
    
    tdenv.DEBUG0("Marking populated stations as having a market")
    db.execute(
        "UPDATE Station SET market = 'Y'"
        " WHERE EXISTS"
            " (SELECT station_id FROM StationItem"
              " WHERE StationItem.station_id = Station.station_id"
             ")"
    )
    
    tdenv.DEBUG0('Committing...')
    db.commit()
    db.close()
    
    changes = " and ".join("{} {}".format(v, k) for k, v in {
        "new": newItems,
        "updated": updtItems,
        "removed": removedItems,
    }.items() if v) or "0"
    
    tdenv.NOTE(
        "Import complete: "
            "{:s} items "
            "over {:n} stations "
            "in {:n} systems",
                changes,
                len(stations),
                numSys,
    )
    
    if ignItems:
        tdenv.NOTE("Ignored {} items with old data", ignItems)


######################################################################


def depCheck(importPath, lineNo, depType, key, correctKey):
    if correctKey == key:
        return
    if correctKey == corrections.DELETED:
        raise DeletedKeyError(importPath, lineNo, depType, key)
    raise DeprecatedKeyError(importPath, lineNo, depType, key, correctKey)


def deprecationCheckSystem(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'System',
        line[0], corrections.correctSystem(line[0]),
    )


def deprecationCheckStation(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'System',
        line[0], corrections.correctSystem(line[0]),
    )
    depCheck(
        importPath, lineNo, 'Station',
        line[1], corrections.correctStation(line[0], line[1]),
    )


def deprecationCheckCategory(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'Category',
        line[0], corrections.correctCategory(line[0]),
    )


def deprecationCheckItem(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'Category',
        line[0], corrections.correctCategory(line[0]),
    )
    depCheck(
        importPath, lineNo, 'Item',
        line[1], corrections.correctItem(line[1]),
    )


def processImportFile(tdenv, db, importPath, tableName, *, line_callback: Optional[Callable] = None, call_args: Optional[dict] = None):
    tdenv.DEBUG0(
        "Processing import file '{}' for table '{}'",
        str(importPath), tableName
    )
    call_args = call_args or {}
    if line_callback:
        line_callback = partial_fn(line_callback, **call_args)
    
    fkeySelectStr = (
        "("
        " SELECT {newValue}"
        " FROM {table}"
        " WHERE {stmt}"
        ")"
    )
    uniquePfx = "unq:"
    uniqueLen = len(uniquePfx)
    ignorePfx = "!"
    
    with importPath.open('r', encoding='utf-8') as importFile:
        csvin = csv.reader(
            importFile, delimiter=',', quotechar="'", doublequote=True
        )
        # first line must be the column names
        columnDefs = next(csvin)
        columnCount = len(columnDefs)
        
        # split up columns and values
        # this is necessary because the insert might use a foreign key
        bindColumns = []
        bindValues = []
        joinHelper = []
        uniqueIndexes = []
        for (cIndex, cName) in enumerate(columnDefs):
            colName, _, srcKey = cName.partition('@')
            # is this a unique index?
            if colName.startswith(uniquePfx):
                uniqueIndexes.append(cIndex)
                colName = colName[uniqueLen:]
            if not srcKey:
                # no foreign key, straight insert
                bindColumns.append(colName)
                bindValues.append('?')
                continue
            
            queryTab, _, queryCol = srcKey.partition('.')
            if colName.startswith(ignorePfx):
                # this column is only used to resolve an FK
                assert srcKey
                colName = colName[len(ignorePfx):]
                joinHelper.append((colName, queryTab, queryCol))
                continue
            
            # foreign key, we need to make a select
            joinTable = [ queryTab ]
            joinStmt = []
            for nextCol, nextTab, nextJoin in joinHelper:
                joinTable.append(
                    "INNER JOIN {} USING({})".format(nextTab, nextJoin)
                )
                joinStmt.append(
                    "{}.{} = ?".format(nextTab, nextCol)
                )
            joinHelper = []
            joinStmt.append("{}.{} = ?".format(queryTab, colName))
            bindColumns.append(queryCol)
            bindValues.append(
                fkeySelectStr.format(
                    newValue = srcKey,
                    table = " ".join(joinTable),
                    stmt = " AND ".join(joinStmt),
                )
            )
        # now we can make the sql statement
        sql_stmt = """
            INSERT OR REPLACE INTO {table} ({columns}) VALUES({values})
        """.format(
            table=tableName,
            columns=','.join(bindColumns),
            values=','.join(bindValues)
        )
        tdenv.DEBUG0("SQL-Statement: {}", sql_stmt)
        
        # Check if there is a deprecation check for this table.
        deprecationFn = getattr(
            sys.modules[__name__],
            "deprecationCheck" + tableName,
            None
        )
        
        # import the data
        importCount = 0
        uniqueIndex = {}
        
        for linein in csvin:
            if line_callback:
                line_callback()
            if not linein:
                continue
            lineNo = csvin.line_num
            if len(linein) == columnCount:
                tdenv.DEBUG1("       Values: {}", ', '.join(linein))
                if deprecationFn:
                    try:
                        deprecationFn(importPath, lineNo, linein)
                    except (DeprecatedKeyError, DeletedKeyError) as e:
                        if not tdenv.ignoreUnknown:
                            raise e
                        e.category = "WARNING"
                        tdenv.NOTE("{}", e)
                        continue
                if uniqueIndexes:
                    # Need to construct the actual unique index key as
                    # something less likely to collide with manmade
                    # values when it's a compound.
                    keyValues = [
                        str(linein[col]).upper()
                        for col in uniqueIndexes
                    ]
                    key = ":!:".join(keyValues)
                    prevLineNo = uniqueIndex.get(key, 0)
                    if prevLineNo:
                        # Make a human-readable key
                        key = "/".join(keyValues)
                        raise DuplicateKeyError(
                            importPath, lineNo,
                            "entry", key,
                            prevLineNo
                        )
                    uniqueIndex[key] = lineNo
                
                try:
                    db.execute(sql_stmt, linein)
                    importCount += 1
                except Exception as e:  # pylint: disable=broad-exception-caught
                    tdenv.WARN(
                        "*** INTERNAL ERROR: {err}\n"
                        "CSV File: {file}:{line}\n"
                        "SQL Query: {query}\n"
                        "Params: {params}\n"
                        .format(
                            err = str(e),
                            file = str(importPath),
                            line = lineNo,
                            query = sql_stmt.strip(),
                            params = linein
                        )
                    )
                    pass
            else:
                tdenv.NOTE(
                        "Wrong number of columns ({}:{}): {}",
                            importPath,
                            lineNo,
                            ', '.join(linein)
                )
        db.commit()
        tdenv.DEBUG0("{count} {table}s imported",
                            count = importCount,
                            table = tableName)

######################################################################


def buildCache(tdb, tdenv):
    """
    Rebuilds the SQlite database from source files.
    
    TD's data is either "stable" - information that rarely changes like Ship
    details, star systems etc - and "volatile" - pricing information, etc.
    
    The stable data starts out in data/TradeDangerous.sql while other data
    is stored in custom-formatted text files, e.g. ./TradeDangerous.prices.
    
    We load both sets of data into an SQLite database, after which we can
    avoid the text-processing overhead by simply checking if the text files
    are newer than the database.
    """
    
    tdenv.NOTE(
        "Rebuilding cache file: this may take a few moments.",
        stderr=True,
    )
    
    dbPath = tdb.dbPath
    sqlPath = tdb.sqlPath
    pricesPath = tdb.pricesPath
    
    # Create an in-memory database to populate with our data.
    tempPath = dbPath.with_suffix(".new")
    backupPath = dbPath.with_suffix(".old")
    
    if tempPath.exists():
        tempPath.unlink()
    
    tempDB = sqlite3.connect(str(tempPath))
    tempDB.execute("PRAGMA foreign_keys=ON")
    # Read the SQL script so we are ready to populate structure, etc.
    tdenv.DEBUG0("Executing SQL Script '{}' from '{}'", sqlPath, os.getcwd())
    with sqlPath.open('r', encoding = 'utf-8') as sqlFile:
        sqlScript = sqlFile.read()
        tempDB.executescript(sqlScript)
    
    # import standard tables
    with Progress(max_value=len(tdb.importTables) + 1, prefix="Importing", width=25, style=CountingBar) as prog:
        for importName, importTable in tdb.importTables:
            import_path = Path(importName)
            import_lines = file_line_count(import_path, missing_ok=True)
            with prog.sub_task(max_value=import_lines, description=importTable) as child:
                prog.increment(value=1)
                call_args = {'task': child, 'advance': 1}
                try:
                    processImportFile(tdenv, tempDB, import_path, importTable, line_callback=prog.update_task, call_args=call_args)
                except FileNotFoundError:
                    tdenv.DEBUG0(
                        "WARNING: processImportFile found no {} file", importName
                    )
                except StopIteration:
                    tdenv.NOTE(
                        "{} exists but is empty. "
                        "Remove it or add the column definition line.",
                        importName
                    )
        prog.increment(1)
        
        with prog.sub_task(description="Save DB"):
            tempDB.commit()
    
    # Parse the prices file
    if pricesPath.exists():
        with Progress(max_value=None, width=25, prefix="Processing prices file"):
            processPricesFile(tdenv, tempDB, pricesPath)
    else:
        tdenv.NOTE(
                "Missing \"{}\" file - no price data.",
                    pricesPath,
                    stderr=True,
        )
        tempDB.close()
    
    tdb.close()
    
    tdenv.DEBUG0("Swapping out db files")
    
    if dbPath.exists():
        if backupPath.exists():
            backupPath.unlink()
        dbPath.rename(backupPath)
    tempPath.rename(dbPath)
    
    tdenv.DEBUG0("Finished")

######################################################################


def regeneratePricesFile(tdb, tdenv):
    tdenv.DEBUG0("Regenerating .prices file")
    
    with tdb.pricesPath.open("w", encoding = 'utf-8') as pricesFile:
        prices.dumpPrices(
                tdb.dbFilename,
                prices.Element.full,
                file = pricesFile,
                debug = tdenv.debug)
    
    # Update the DB file so we don't regenerate it.
    os.utime(tdb.dbFilename)

######################################################################


def importDataFromFile(tdb, tdenv, path, pricesFh = None, reset = False):
    """
        Import price data from a file on a per-station basis,
        that is when a new station is encountered, delete any
        existing records for that station in the database.
    """
    
    if not pricesFh and not path.exists():
        raise TradeException("No such file: {}".format(
                    str(path)
                ))
    
    if reset:
        tdenv.DEBUG0("Resetting price data")
        with tdb.getDB() as db:
            db.execute("DELETE FROM StationItem")
            db.commit()
    
    tdenv.DEBUG0("Importing data from {}".format(str(path)))
    processPricesFile(tdenv,
            db = tdb.getDB(),
            pricesPath = path,
            pricesFh = pricesFh,
    )
    
    # If everything worked, we may need to re-build the prices file.
    if path != tdb.pricesPath:
        regeneratePricesFile(tdb, tdenv)



"""
Provides a library of mechanisms for formatting output text to ensure consistency across
TradeDangerous and plugin tools,
"""
from __future__ import annotations

import itertools
import typing

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Optional  # noqa


class ColumnFormat:
    """
        Describes formatting of a column to be populated with data.
        
        Member Functions:
            
            text()
                Applies all formatting (except qualifier) to the name to
                produce a correctly sized title field.
            
            format(value)
                Applies all formatting to key(value) to produce a correctly
                sized value field.
        
        Attributes:
            name
                Heading for column to display when calling title()
                e.g. name="Station Name"
            align
                Alignment formatter for .format
                e.g. align='<' or align='>' or align=''
            width
                Numeric value for the width of the column
                e.g. width=5
            qualifier
                Final part of the print format,
                e.g. qualifier='.2f' or qualifier='n'
            pre
                Prefix to the column
            post
                Postfix to the column
            key
                Retrieve the printable name of the item
            pred
                Predicate: Return False to leave this column blank
        
        e.g.
            cols = [
                ColumnFormat("Name", "<", "5", '', key=lambda item:item['name']),
                ColumnFormat("Dist", ">",  "6", ".2f",
                        pre='[',
                        post=']'
                        key=lambda item:item['dist']),
            ]
            rows = [ {'name':'Bob', 'dist':1.5}, {'name':'John', 'dist':23}]
            # print titles
            print(*[col.text() for col in cols])
            for row in rows:
                print(*[col.format(row) for col in cols])
        Produces:
            Name   [ Dist]
            Bob    [ 1.30]
            John   [23.00]
    
    """
    name:       str                 # name of the column
    align:      str                 # format's alignment specifier
    width:      int                 # width specifier
    qualifier:  Optional[str]       # optional format type specifier e.g. '.2f', 's', 'n'
    pre:        Optional[str]       # prefix to the column
    post:       Optional[str]       # postfix to the column
    key:        Callable            # function to retrieve the printable name of the item
    pred:       Callable            # predicate: return False to leave this column blank
    
    def __init__(
            self,
            name,
            align,
            width,
            qualifier=None,
            pre=None,
            post=None,
            key=lambda item: item,
            pred=lambda item: True,
            ) -> None:
        self.name = name
        self.align = align
        self.width = max(int(width), len(name))
        self.qualifier = qualifier or ''
        self.key = key
        self.pre = pre or ''
        self.post = post or ''
        self.pred = pred
    
    def __str__(self) -> str:
        return f'{self.pre}{self.name:{self.align}{self.width}}{self.post}'
    text = __str__
    
    def format(self, value: str) -> str:
        """ Returns the string formatted with a specific value"""
        if not self.pred(value):
            return f'{self.pre}{"":{self.align}{self.width}}{self.post}'
        return f'{self.pre}{self.key(value):{self.align}{self.width}{self.qualifier}}{self.post}'

class RowFormat:
    """
        Describes an ordered collection of ColumnFormats
        for dispay data from rows, such that calling
          rowFmt.format(rowData)
        will return the result of formatting each column
        against rowData.
        
        Member Functions
            
            append(col, [after])
                Adds a ColumnFormatter to the end of the row
                If 'after' is specified, tries to insert
                the new column immediately after the first
                column who's name matches after.
            
            insert(pos, newCol)
                Inserts a ColumnFormatter at position pos in the list
            
            text()
                Returns a list of all the column headings
            
            format(rowData):
                Returns a list of applying rowData to all
                of the columns
    
    """
    columns:        list[ColumnFormat]
    prefix:         str
    
    def __init__(self, prefix: Optional[str] = None):
        self.columns = []
        self.prefix = prefix or ""
    
    def addColumn(self, *args, **kwargs) -> None:
        self.append(ColumnFormat(*args, **kwargs))
    
    def append(self, column: ColumnFormat, after: Optional[str] = None) -> 'RowFormat':
        columns = self.columns
        if after:
            for idx, col in enumerate(columns, 1):
                if col.name == after:
                    columns.insert(idx, column)
                    return self
        columns.append(column)
        return self
    
    def insert(self, pos: int, column: Optional[ColumnFormat]) -> None:
        if column is not None:
            self.columns.insert(pos, column)
    
    def __str__(self) -> str:
        return f"{self.prefix} {' '.join(str(col) for col in self.columns)}"
    
    text = __str__  # alias
    
    def heading(self) -> tuple[str, str]:
        """ Returns a title and the appropriate underline for that text. """
        headline = f"{self}"
        return headline, '-' * len(headline)
    
    def format(self, row_data: Optional[Any]) -> str:
        return f"{self.prefix} {' '.join(col.format(row_data) for col in self.columns)}"

def max_len(iterable, key=lambda item: item):
    iterable, readahead = itertools.tee(iter(iterable))
    try:
        next(readahead)
    except StopIteration:
        return 0
    return max(len(key(item)) for item in iterable)


if __name__ == '__main__':
    rowFmt = RowFormat(). \
                append(ColumnFormat("Name", '<', '8', key=lambda row: row['name'])). \
                append(ColumnFormat("Dist", '>', '6', '.2f', pre='[', post=']', key=lambda row: row['dist']))
    
    rows = [
        { 'name': 'Bob', 'dist': 6.2, 'age': 30 },
        { 'name': 'Dave', 'dist': 42, 'age': 18 },
    ]
    
    def present():
        rowTitle = rowFmt.text()
        print(rowTitle)
        print('-' * len(rowTitle))
        for row in rows:
            print(rowFmt.format(row))
    
    print("Simple usage:")
    present()
    
    # print()
    # print("Adding age ColumnFormat:")
    # rowFmt.append(after='Name', col=ColumnFormat("Age", '>', 3, pre='|', post='|', key=lambda row: row['age']))
    # present()
